#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <string.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <unistd.h>

#include <errno.h>

#define TYPE_ADD	0x01
#define TYPE_DEL	0x02
#define TYPE_GET	0x03

#define BUF_REQUEST	128

#define SOCK_PATH "/tmp/pidentd.sock"

typedef struct {
	unsigned char	querytype;
	in_port_t 	localport;
	in_port_t 	remoteport;
	in_addr_t 	localaddr;
	in_addr_t 	remoteaddr;
	pid_t		pid;
	uid_t		uid;
	int		fd;
} chain_t;

chain_t globtab[BUF_REQUEST];

void pidentd_add(chain_t tmp);
void pidentd_del(chain_t tmp);
uid_t pidentd_get(chain_t tmp);


void
pidentd_add(chain_t tmp)
{
	int firstfree = -1;
	int i;

	/* look after a pre-existing localport */ 
	for(i = 0; i < BUF_REQUEST; i ++) {
		if(globtab[i].querytype != 0
		&& globtab[i].localport == tmp.localport)
			break;
		if(firstfree == -1 && globtab[i].querytype == 0)
			firstfree = i;
	}

#ifdef DEBUG
	fprintf(stderr, "add: %d -> %d\n", tmp.localport, tmp.remoteport);
#endif

	if(i == BUF_REQUEST && firstfree == -1) { /* arg. */
		return;
	} else if(i == BUF_REQUEST) { /* fill firstfree */ 
		memcpy(&globtab[firstfree], &tmp, sizeof(chain_t)); 
	} else { /* fill i */
		memcpy(&globtab[i], &tmp, sizeof(chain_t));
	}

	return;
}

void
pidentd_del(chain_t tmp)
{
	int i;

	/* look after a pre-existing localport */
	for(i = 0; i < BUF_REQUEST; i ++) {
		if(globtab[i].querytype != 0
		&& globtab[i].localport == tmp.localport) 
			break;
	}

	if(i != BUF_REQUEST) {
		globtab[i].querytype = 0;
	}

	return;
}

uid_t
pidentd_get(chain_t tmp)
{
	int i;

	/* look after a existing localport */
	for(i = 0; i < BUF_REQUEST; i ++) {
		if(globtab[i].querytype != 0
		&& globtab[i].localport == tmp.localport)
			break;
#ifdef DEBUG
		else {
			if(globtab[i].querytype != 0)
				fprintf(stderr, "%d %d\n", globtab[i].localport, tmp.localport);
		}
#endif
	}

	if(i != BUF_REQUEST) {
		return(globtab[i].uid);
	}

	return(-1);
}

int
main(void)
{
	int sockfd, ret, fdc;
	struct sockaddr_un ad;
	uid_t t;
	size_t len;

	fd_set pom;

	chain_t s;

	ret = unlink(SOCK_PATH);
	if(ret == -1) {
		if(errno != ENOENT) {
			perror("unlink");
			exit(-1);
		}
	}

	sockfd = socket(PF_LOCAL, SOCK_STREAM, 0);

	if(sockfd == -1) {
		perror("socket");
		exit(-1);
	}

	bzero(&ad, sizeof(struct sockaddr_un));

	ad.sun_family = PF_LOCAL;
	strncpy(ad.sun_path, SOCK_PATH, sizeof(ad.sun_path) - 1); 

	ret = bind(sockfd, (struct sockaddr*)&ad, sizeof(ad));
	if(ret == -1) {
		perror("bind");
		exit(-1);
	}

	FD_SET(sockfd, &pom);
	len = sizeof(ad); 

	ret = listen(sockfd, 16);
	if(ret == -1) {
		perror("listen");
		exit(-1);
	}

#ifndef DEBUG
	if(fork()) {
		exit(0);
	}
	setsid();
#endif

	while((fdc = accept(sockfd, (struct sockaddr*)&ad, &len)) != -1) {
#ifdef DEBUG
		perror("accept");
		fprintf(stderr, "request!\n");
#endif
		ret = read(fdc, &s, sizeof(s));

#ifdef DEBUG
		perror("read");
		fprintf(stderr, "%d %d\n", ret, sizeof(s));
#endif

		if(ret != sizeof(s)) {
			close(fdc);
			continue;
		}

		switch(s.querytype) {
			case TYPE_ADD:
				pidentd_add(s);
			break;
			case TYPE_DEL:
				pidentd_del(s);
			break;
			case TYPE_GET:
#ifdef DEBUG
				fprintf(stderr, "look for: %d\n", s.localport);
#endif
				t = pidentd_get(s);
				ret = write(fdc, &t, sizeof(uid_t)); 
				usleep(50000);
			break;
			default:
				goto end;
			break;
		}

		close(fdc);

	} 

end:
	/* error */
#ifdef DEBUG
	fprintf(stderr, "Error occured. Exiting.\n");
#endif
	close(sockfd);

	return -1;
}
