#include <dlfcn.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#define	TYPE_ADD	0x01
#define	TYPE_DEL	0x02

#define	SOCK_PATH	"/tmp/pidentd.sock"

typedef struct {
	unsigned char	querytype;
	in_port_t	localport;
	in_port_t	remoteport;
	in_addr_t	localaddr;
	in_addr_t	remoteaddr;
	pid_t	pid;
	uid_t	uid;
	int	fd;
} chain_t;


int (*old_connect)(int s, const struct sockaddr *name, socklen_t namelen) = NULL; 
int (*old_close)(int s) = NULL;

int getlibcconnectaddr(void);
void add_clue(in_port_t localport, in_port_t remoteport, in_addr_t localaddr, in_addr_t remoteaddr, int fd);
void del_clue(in_port_t localport, int fd);


void
add_clue(in_port_t localport,
         in_port_t remoteport,
         in_addr_t localaddr,
         in_addr_t remoteaddr,
         int fd)
{
	int sockfd, res;
	chain_t p;
	struct sockaddr_un ad;

	p.querytype = TYPE_ADD;
	p.localport = localport;
	p.remoteport = remoteport;
	p.localaddr = localaddr;
	p.remoteaddr = remoteaddr;
	p.pid = getpid();
	p.uid = getuid();
	p.fd = fd;

	sockfd = res = 0;
	bzero(&ad, sizeof(ad));

	sockfd = socket(PF_LOCAL, SOCK_STREAM, 0); 
	ad.sun_family = PF_LOCAL;
	strncpy(ad.sun_path, SOCK_PATH, sizeof(ad.sun_path) - 1); 

	res = old_connect(sockfd, (struct sockaddr*)&ad, sizeof(ad));

	if(res != -1) {
		write(sockfd, &p, sizeof(chain_t));
	}

	old_close(sockfd);

	return;
}

void
del_clue(in_port_t localport, int fd)
{
	int sockfd, res;
	chain_t p;
	struct sockaddr_un ad;

	p.querytype = TYPE_DEL;
	p.localport = localport;
	p.fd = fd;

	bzero(&p, sizeof(chain_t));

	sockfd = socket(PF_LOCAL, SOCK_STREAM, 0);
	ad.sun_family = PF_LOCAL;
	strncpy(ad.sun_path, SOCK_PATH, sizeof(ad.sun_path) - 1);

	res = old_connect(sockfd, (struct sockaddr*)&ad, sizeof(ad));

	if(res != -1) {
		write(sockfd, &p, sizeof(chain_t));
	}

	old_close(sockfd);

	return;
}

int
getlibcconnectaddr(void)
{
	void *res;

	if ((res = dlsym(RTLD_NEXT, "connect")) == NULL) {
#ifdef DEBUG
		fprintf(stderr, "dlsym %s error:%s\n", "connect", dlerror());
#endif
		exit(-1);
	}

	old_connect = res;

	if ((res = dlsym(RTLD_NEXT, "close")) == NULL) {
#ifdef DEBUG
		fprintf(stderr, "dlsym %s error:%s\n", "close", dlerror());
#endif
		exit(-1);
	}

	old_close = res;

	return 0;
}

int
connect(int s, const struct sockaddr *name, socklen_t namelen)
{
	int resgpn, resgsn, resc;
	struct sockaddr_in sagpn, sagsn;
	socklen_t slgpn, slgsn;

	if(old_connect == NULL) {
		getlibcconnectaddr();
	}

#ifdef DEBUG
	printf("connect from pid: %d (user: %d) !\n", getpid(), getuid());
#endif

	resc = old_connect(s, name, namelen);

	if(resc != -1 && name != NULL) { /* try a getpeername to get remote port */
		resgpn = getpeername(s, (struct sockaddr*)&sagpn, &slgpn); 

		if(resgpn == -1) {
#ifdef DEBUG
			perror("getpeername");
			return(resc);
#endif
		}

		if(name->sa_family != PF_INET) {
#ifdef DEBUG
			fprintf(stderr, "not a inet socket (%d).\n",
			        name->sa_family);
#endif
			return(resc);
		}

		resgsn = getsockname(s, (struct sockaddr*)&sagsn, &slgsn);

		if(resgpn != -1) {
#ifdef DEBUG
			printf("port: %d -> %d\n",
			       ntohs(sagpn.sin_port), 
			       ntohs(sagsn.sin_port));
#endif

			add_clue(ntohs(sagsn.sin_port),
			         ntohs(sagpn.sin_port),
			         ntohl(sagsn.sin_addr.s_addr),
			         ntohl(sagpn.sin_addr.s_addr),
			         resc);
		} else {
#ifdef DEBUG
			perror("getsockname");
#endif
		}
	}	

	return(resc);
}

int
close(int s)
{
	int resgsn;
	socklen_t slgsn;
	struct sockaddr_in sagsn;

	if(old_close == NULL) {
		getlibcconnectaddr();
	}

#ifdef DEBUG
        printf("close from pid: %d (user: %d) %d !\n", getpid(), getuid(), s);
#endif  

	resgsn = getsockname(s, (struct sockaddr*)&sagsn, &slgsn);

	if(resgsn != -1 && sagsn.sin_family == PF_INET) {
		del_clue(ntohs(sagsn.sin_port), s);
	}

	return(old_close(s));
}
